import copy
import logging
import traceback
from typing import Any, Dict, List, Optional

from nornir.core.task import Result, Task

from .connections import CONNECTION_NAME

GetterOptionsDict = Optional[Dict[str, Dict[str, Any]]]
from .. import get_network_driver

logger = logging.getLogger(__name__)


def umnet_napalm_get(
    task: Task,
    getters: List[str],
    getters_options: GetterOptionsDict = None,
    **kwargs: Any
) -> Result:
    """
    Gather information from network devices using napalm

    Arguments:
        getters: getters to use
        getters_options (dict of dicts): When passing multiple getters you
            pass a dictionary where the outer key is the getter name
            and the included dictionary represents the options to pass
            to the getter
        **kwargs: will be passed as they are to the getters

    Returns:
        Result object with the following attributes set:
          * result (``dict``): dictionary with the result of the getter
    """
    try:
        device = task.host.get_connection(CONNECTION_NAME, task.nornir.config)
        getters_options = getters_options or {}

        if isinstance(getters, str):
            getters = [getters]

        result = {}
        for g in getters:
            options = copy.deepcopy(kwargs)
            options.update(getters_options.get(g, {}))
            getter = g if g.startswith("get_") else "get_{}".format(g)
            method = getattr(device, getter)
            result[g] = method(**options)

    except Exception as e:
        tb = traceback.format_exc()
        logger.error(
            "Host %r getters failed with traceback:\n%s",
            task.host.name,
            tb,
        )
        return Result(host=task.host, result=tb, exception=str(e), failed=True)
    return Result(host=task.host, result=result)


def umnet_napalm_get_in_context(
    task: Task,
    getters: Dict[str, GetterOptionsDict],
):
    """
    This task runs the list of getters in a napalm context manager -
    opening a connection, running the getters, then closing the connection.

    Find that this is more resilient, particularly in a multiprocess scenario.
    """
    parameters = {
        "hostname": task.host.hostname,
        "username": task.host.username,
        "password": task.host.password,
        "optional_args": {},
    }
    parameters.update(task.host.connection_options["umnet_napalm"].extras)

    network_driver = get_network_driver(task.host.platform)

    result = {}
    try:
        with network_driver(**parameters) as device:

            for getter, options in getters.items():
                method = getattr(device, getter)
                result[getter] = method(**options)

        return Result(host=task.host, result=result)
    except:
        return Result(host=task.host, result="Connection error")



# # pylint: disable=no-member
from typing import List, Union, Optional, Dict
import re
from ipaddress import ip_network
import logging

from napalm.junos import JunOSDriver
from napalm.base.models import ARPTableDict

from lxml import etree

from ..base import UMnetNapalmError, UMnetNapalm
from ..models import (
    RouteDict,
    MPLSDict,
    InventoryDict,
    IPInterfaceDict,
    LagInterfaceDict,
)
from ..utils import age_to_integer

from . import junos_views

logger = logging.getLogger(__name__)

XCVR_DESC_TO_INTERFACE_PREFIX = {
    r"QSFP": "et-",
    r"SFP\+": "xe-",
    r"SFP-": "ge-",
    r"CFP-": "et-",
    r"UNKNOWN": "ge-",  # default
}

DEFAULT_ARP_AGING_TIMER = 1200


class Junos(UMnetNapalm, JunOSDriver):
    """
    Junos parser
    """

    # interfaces to ignore when doing 'show ip interfaces'
    IGNORE_INTERFACES = [
        "bme",
        "jsrv",
        "em1",
        "em0",
    ]

    # tables to ignore when doing 'show route'
    IGNORE_ROUTE_TABLES = [
        "bgp.evpn.0",
        "bgp.l3vpn.0",
        "bgp.l3vpn-inet6.0",
        "inet.1",
        "inet.2",
        "inet.3",
        "inet.4",
        "inet6.3",
        "inet6.1",
        "iso.0",
        "mpls.0",
        ":vxlan.inet.0",
        "__default_evpn__.evpn.0",
        "default-switch.evpn.0",
    ]

    # mapping of inventory name regex to their netbox types
    INVENTORY_TO_TYPE = {
        r"Fan Tray": "fan",
        r"(PIC|MIC|FPC)": "linecard",
        r"Xcvr": "optic",
        r"(PEM|Power Supply)": "psu",
        r"Routing Engine": "re",
        # don't care about these inventory items
        r"(CPU|PDM|FPM|Midplane|CB)": None,
    }

    def __init__(
        self,
        hostname: str,
        username: str,
        password: str,
        timeout: int = 60,
        optional_args: Optional[Dict] = None,
    ) -> None:

        super().__init__(hostname, username, password, timeout, optional_args)

        self._running_config = None
        self.arp_aging_timer = DEFAULT_ARP_AGING_TIMER

    def _get_junos_inventory_type(self, name: str, model_number: str) -> str:
        """
        Maps the name and part number of the Junos inventory item to its type
        """
        # "Xcvr" parts are always optics
        if name.startswith("Xcvr"):
            return "optic"

        # non-optic parts should always have a model number, if not
        # we don't care about them
        if not (model_number):
            return None

        # EX switches and virtual chassis save their vc members and uplink modules
        # as FPC X and PIC X - we want to classify those correctly
        if re.match(r"EX[234]", model_number):
            if re.search(r"FPC \d+", name):
                return "stack-member"

            if re.search(r"PIC \d+", name):
                return None

        # uplink modules are also saved under PIC X, must classify based
        # on their model number
        if re.match(r"EX-UM", model_number):
            return "uplink-module"

        # otherwise we want to pattern-match based on the INVENTORY_TO_TYPE
        # dictionary
        return self._get_inventory_type(name)

    def _get_xcvr_interface_prefix(self, xcvr_desc: str) -> str:
        """
        Maps xcvr description from "show chassis hardware" to
        interface prefix
        """
        for pattern, prefix in XCVR_DESC_TO_INTERFACE_PREFIX.items():
            if re.match(pattern, xcvr_desc):
                return prefix

        raise UMnetNapalmError(f"{self.hostname}: Unknown xcvr type {xcvr_desc}")

    def _get_mod_number(self, mod_name: str) -> str:
        """
        Strips out "FPC|MIC|PIC" from a module name seen in "show chassis hardware"
        """
        return re.sub(r"(MIC|PIC|FPC) ", "", mod_name)

    def _send_command(self, view: str) -> str:
        """
        Uses junos views to send an RPC command and gets the results
        """
        view = getattr(junos_views, view)
        result = view(self.device)  # pylint: disable=no-member
        return result.get()

    def _parse_dhcp_helpers(self):
        """
        Parses dhcp relay helpers from the running config
        """

        server_groups = {"default": {}}
        groups = {"default": {}}

        vrf = "default"
        group = None

        dhcp_relay_flag = False
        dhcp_relay_indent = ""
        vrf_flag = False
        server_group_flag = False
        indent = ""

        for line in self.running_config.split("\n"):

            # detecting which vrf we're in
            m = re.match(r"routing-instances {", line)
            if m:
                vrf_flag = True
                continue

            m = re.match(r"    (\S+) {", line)
            if m and vrf_flag:
                vrf = m.group(1)
                server_groups[vrf] = {}
                groups[vrf] = {}
                continue

            m = re.match(r"}$", line)
            if m and vrf_flag:
                vrf_flag = False
                continue

            # dhcp relay section flag
            m = re.match(r"(\s+)dhcp-relay {", line)
            if m:
                dhcp_relay_flag = True
                dhcp_relay_indent = m.group(1)
                continue

            # dhcp server group parsing start
            m = re.match(r"(\s+)server-group {", line)
            if m and dhcp_relay_flag:
                indent = m.group(1)
                server_group_flag = True
                continue

            # new server group definition
            m = re.match(r"\s+(\S+) {", line)
            if m and dhcp_relay_flag and server_group_flag:
                server_group = m.group(1)
                server_groups[vrf][server_group] = []
                continue

            # new server in server group
            m = re.match(r"\s+(\d+.\d+.\d+.\d+);", line)
            if m and dhcp_relay_flag and server_group_flag:
                server_groups[vrf][server_group].append(m.group(1))
                continue

            ### dhcp group parsing start
            m = re.match(r"(\s+)group (\S+) {", line)
            if m and dhcp_relay_flag:
                indent = m.group(1)
                group = m.group(2)
                groups[vrf][group] = {"interfaces": []}
                continue

            # active-server group for group
            m = re.match(r"\s+active-server-group (\S+);", line)
            if m and dhcp_relay_flag and group:
                groups[vrf][group]["server-group"] = m.group(1)
                continue

            # interface in group
            m = re.match(r"\s+interface (\S+);", line)
            if m and dhcp_relay_flag and group:
                groups[vrf][group]["interfaces"].append(m.group(1))
                continue

            # end server group or group parsing
            m = re.match(r"(\s+)}", line)
            if m and m.group(1) == indent and server_group_flag:
                server_group_flag = False
                continue

            if m and m.group(1) == indent and group:
                group = False
                continue

            if m and m.group(1) == dhcp_relay_indent:
                dhcp_relay_flag = False
                continue

        # final output 'resolves' the server-group name into
        # list of IPs from the server-group dictionary
        output = {}
        for vrf, groups in groups.items():
            output[vrf] = {}
            for group in groups.values():
                if (
                    vrf not in server_groups
                    or group["server-group"] not in server_groups[vrf]
                ):
                    continue
                server_list = server_groups[vrf][group["server-group"]]
                for interface in group["interfaces"]:
                    output[vrf][interface] = server_list

        return output

    @property
    def running_config(self):
        if not self._running_config:
            self.get_config(retrieve="running")
        return self._running_config

    def get_config(self, retrieve="all", full=False, sanitized=False):
        """
        If we run our nornir getters right we can pull the arp aging timer
        out of the config
        """
        config = super().get_config(retrieve, full, sanitized)
        if "running" in config:
            self._running_config = config["running"]
        return config

    def get_arp_table(self, vrf="") -> ARPTableDict:
        """
        Pulls ARP table - napalm does have get_arp_table for junos but it doesn't
        support VRFs :(

        Because we have to dynamically query every VPN on the device we can't use
        junos views for this since they're based on statically defined rpc calls.
        Instead we'll have to do our own rpc calls.
        """

        # figure out the ARP aging timer - no operational command appears to show this,
        # pulling it from the config.
        m = re.search(r"system \{.+arp {\+aging-timer (\d+);", self.running_config)
        if m:
            self.arp_aging_timer = int(m.group(1))

        # building the rpc call as an lxml element so we can dynamically add a vrf if we want
        # if no vrf is provided all ARP entries are returned
        arp_rpc = etree.Element("get-arp-table-information")
        etree.SubElement(arp_rpc, "no-resolve")
        etree.SubElement(arp_rpc, "expiration-time")
        if vrf:
            xml_vrf = etree.SubElement(arp_rpc, "vpn")
            xml_vrf.text = vrf

        rpc_result = self.device.execute(arp_rpc)

        results = []
        for xml_entry in rpc_result.iterfind("arp-table-entry"):

            interface = xml_entry.find("interface-name").text.strip("\n")

            # MX and EX9200 include the physical interface in brackets - stripping
            # that out so only the logical interface is shown
            interface = re.sub(r" \[.+\]$", "", interface)

            exp = xml_entry.find("time-to-expire")
            if exp is not None:
                age = self.arp_aging_timer - int(exp.text)
            else:
                age = 0

            results.append(
                {
                    "interface": interface,
                    "ip": xml_entry.find("ip-address").text.strip("\n"),
                    "mac": xml_entry.find("mac-address").text.strip("\n").upper(),
                    "age": age,
                }
            )

        return results

    def get_ip_interfaces(self) -> List[IPInterfaceDict]:
        """
        Parses 'get interface' and show route instance detail
        """

        # parsing dhcp helpers from the running config
        helpers = self._parse_dhcp_helpers()

        interface_vrfs = self._send_command("junos_vrf_interfaces")
        i_vrfs = {i.name: i.vrf for i in dict(interface_vrfs).values()}

        interfaces = self._send_command("junos_ip_interfaces")

        results = []
        for i in dict(interfaces).values():

            if self._ignore_interface(i.interface) or i.family not in ["inet", "inet6"]:
                continue

            # ignore ipv6 link local
            if i.family == "inet6" and i.ip.startswith("fe80"):
                continue

            # ignore non-preferred addresses, these appear to be vrrp virtuals
            if not (i.preferred):
                continue

            # host networks don't have 'ifa-destination' (subnet) field
            if i.subnet:
                prefixlen = i.subnet.split("/")[1]
            elif i.family == "inet":
                prefixlen = "32"
            else:
                prefixlen = "128"

            vrf = i_vrfs.get(i.interface, "default")

            # junos has weird internal private VRFs with loopbacks
            if vrf.startswith("__"):
                continue

            if helpers.get(vrf):
                i_helpers = helpers[vrf].get(i.interface, [])
            else:
                i_helpers = []

            results.append(
                {
                    "ip_address": f"{i.ip}/{prefixlen}",
                    "interface": i.interface,
                    "description": i.description,
                    "mtu": 9216 if i.mtu == "Unlimited" else int(i.mtu),
                    "admin_up": bool(i.admin_up),
                    "oper_up": not (i.device_down or i.hardware_down),
                    "vrf": vrf,
                    "secondary": not (i.primary),
                    "helpers": i_helpers,
                }
            )

        return results

    def get_active_routes(self) -> List[RouteDict]:
        """
        Parses rpc call show route detail
        """
        logger.debug("getting active routes")
        result = self._send_command("junos_route_table")
        logger.debug("Done getting routes")

        output = []

        # junos views returns data as a list of entries, where each entry
        # is a list of tuples (key, val) from the junos view.
        # You can cast the data to a list of dicts (see get_inventory)
        # but for large results converting to dict is very slow so we're not doing
        # that here.
        for (_, vrf), (_, prefix), (_, rt_entry) in result.values():

            if vrf in self.IGNORE_ROUTE_TABLES:
                continue

            prefix = ip_network(prefix)

            if vrf in ["inet.0", "inet6.0"]:
                vrf = "default"
            else:
                vrf = vrf.split(".")[0]

            # also skip link local prefixes
            if prefix.is_link_local:
                continue

            for (
                (_, age),
                (_, protocol),
                (_, learned_from),
                (_, next_hop),
            ) in rt_entry.values():

                for (
                    (_, nh_ip),
                    (_, nh_interface),
                    (_, mpls_label),
                    (_, selected_next_hop),
                ) in next_hop.values():

                    if not selected_next_hop:
                        continue

                    if learned_from:
                        learned_from = learned_from
                    elif nh_ip:
                        learned_from = nh_ip
                    elif protocol in ["Direct", "Local", "Static"]:
                        learned_from = "self"
                    else:
                        raise UMnetNapalmError(
                            f"Could not determine learned from for {prefix}"
                        )

                    mpls_label = self._parse_mpls_label(mpls_label)

                    output.append(
                        RouteDict(
                            vrf=vrf,
                            prefix=str(prefix),
                            nh_interface=nh_interface,
                            learned_from=learned_from,
                            protocol=protocol,
                            age=age_to_integer(age),
                            nh_ip=nh_ip,
                            mpls_label=mpls_label,
                            nh_table="default" if mpls_label else vrf,
                            # not supporting junos VXLAN
                            vxlan_vni=None,
                            vxlan_endpoint=None,
                        )
                    )

        return output

    def get_inventory(self) -> List[InventoryDict]:
        """
        parses the get-chassis-inventory RPC call, which maps to "show chassis hardware"
        """
        result = self._send_command("junos_inventory")

        output = []
        for chassis in dict(result).values():

            # note that the chassis model and s/n are at this level, but
            # that doesn't count as an 'inventory item' so we're ignoring it

            # saving modules, sub-modules and sub-sub modules
            for module in dict(chassis.modules).values():
                self._save_inventory_item(output, module)

                for sub_module in dict(module.sub_modules).values():
                    self._save_inventory_item(output, sub_module, parent=module.name)

                    for sub_sub_module in dict(sub_module.sub_sub_modules).values():
                        self._save_inventory_item(
                            output,
                            sub_sub_module,
                            parent=sub_module.name,
                            grandparent=module.name,
                        )

        return output

    def _save_inventory_item(
        self, output: list, item: any, parent: str = None, grandparent: str = None
    ) -> bool:
        """
        Extracts data from a particular inventory item object.
        Returns whether we care aobut this item or not (so we know whether or not)
        to loop over its children
        """
        # skip builtin types, or parts without model numbers that aren't transcievers
        if item.part_number == "BUILTIN":
            return False

        # get inventory type based on item name and P/N
        inv_type = self._get_junos_inventory_type(item.name, item.model_number)

        if not inv_type:
            return False

        # for transcievers, change the name from Xcvr X to be the junos interface name
        # note that we expect transcievers to always be at the sub-sub module level
        # and thus to have a grandparent and parent
        m = re.search(r"Xcvr (\d+)", item.name)
        if m:
            if not (grandparent or parent):
                raise UMnetNapalmError(
                    f"{self.hostname}: No MIC and PIC found for {item.name}"
                )
            prefix = self._get_xcvr_interface_prefix(item.description)

            item_name = f"{prefix}{self._get_mod_number(grandparent)}/{self._get_mod_number(parent)}/{m.group(1)}"

        # for sub-linecards (mics or pics) we want to prepend the parent to the item name
        elif parent:
            item_name = f"{parent} {item.name}"
        else:
            item_name = item.name

        output.append(
            {
                "type": inv_type,
                "name": item_name,
                "part_number": self._get_inventory_part_number(item),
                "serial_number": item.serial_number,
            }
        )

        return True

    def _get_inventory_part_number(self, item: any) -> str:
        """
        Extracts the part number from an inventory item
        which, depending on the specific item, is stashed in
        the part number, model number, or description field
        """
        if item.model_number:
            return item.model_number
        return item.description

    def get_mpls_switching(self) -> List[MPLSDict]:
        """
        Parses 'show route active-path table mpls.0'
        """
        result = self._send_command("junos_mpls_table")

        output = []
        for (_, prefix), (_, rt_entry) in result.values():
            # in our network we can skip these 'label is not at the top
            # of the stack' entries - they're redundant
            # also labels 0-2 are junos-internal
            if "(S=0)" in prefix or prefix in ["0", "1", "2"]:
                continue

            for entry in rt_entry.values():
                next_hop = entry[0][1]
                for (
                    (_, nh_ip),
                    (_, nh_interface),
                    (_, mpls_label),
                    (_, nh_table_receive),
                    (_, nh_table),
                ) in next_hop.values():

                    # next hop table is either in 'nh-table-receive' or 'nh-table'
                    nh_table = nh_table if nh_table else nh_table_receive

                    (nh_interface, out_label, aggregate) = self._process_mpls_nh(
                        nh_interface, mpls_label, nh_table
                    )

                    # 'fec' and 'rd' aren't in the Junos mpls.0 table - this is a Cisco
                    # thing from "show mpls switching"
                    output.append(
                        {
                            "in_label": prefix,
                            "out_label": out_label,
                            "nh_interface": nh_interface,
                            "fec": None,
                            "nh_ip": nh_ip,
                            "rd": None,
                            "aggregate": aggregate,
                        },
                    )

        return output

    def _parse_mpls_label(self, label_str) -> Union[List[int], None]:
        """
        Parses mpls-label parameter into a list of integers
        """
        if not label_str:
            return None

        m = re.match(r"Push (?P<inner>\d+)(, Push (?P<outer>\d+))*", label_str)
        if not m:
            raise UMnetNapalmError(f"Couldn't parse label op {label_str}")

        labels = [m.group("inner")]
        if m.group("outer"):
            labels.append(m.group("outer"))

        return labels

    def _process_mpls_nh(
        self, nh_interface: str, mpls_label: str, nh_table: str
    ) -> tuple:
        """
        Parses 'nh interface', 'mpls_label', and nh_table values
        into something our standardized outpur likes.
        Returns a tuple (nh_interface:str, out_label:list, aggregate:bool )
        """
        # when the next hop is a table, not an interface we're going to
        # classify it as an aggregate label
        if not nh_interface and nh_table:
            nh_vrf = re.sub(r".inet6?.0", "", nh_table)
            return (nh_vrf, ["pop"], True)

        # aggregate labels also show up as nh_interface "lsi.X (VRF_NAME)"
        m = re.match(r"^lsi.\d+ \((\S+)\)", nh_interface)
        if m:
            return (m.group(1), [], True)

        # 'pop' label has trailing whitespace
        if mpls_label and re.match(r"^Pop", mpls_label):
            return (nh_interface, ["pop"], False)

        # 'swap' with optional 'push'
        if mpls_label:
            m = re.match(r"^Swap (?P<inner>\d+)(, Push (?P<outer>\d+))*", mpls_label)
            if m:
                out_label = [m.group("inner")]
                if m.group("outer"):
                    out_label.append(m.group("outer"))

                return (nh_interface, out_label, False)

        raise UMnetNapalmError(
            f"Cannot parse MPLS route nh: {nh_interface} label: {mpls_label}"
        )

    def get_lag_interfaces(self) -> Dict[str, LagInterfaceDict]:
        """
        Extracts LAG info from 'show interfaces terse' and 'show lacp interface'
        """
        result = self._send_command("junos_show_int_lag")

        result_dict = dict(result)

        parents = {k: v for k, v in result_dict.items() if k.startswith("ae")}
        children = {k: v for k, v in result_dict.items() if v.lag_name}

        # building parent list
        lag_interfaces = {}
        for name, entry in parents.items():

            # stripping off logical port
            name = name.split(".")[0]

            lag_interfaces[name] = {
                "oper_up": entry.oper_status == "up",
                "admin_up": entry.admin_status == "up",
                "protocol": "Static",  # to be overridden if LAG is seen in 'show lacp interfaces'
                "members": {},
                "mlag_id": 0,  # don't care about MC-LAG (shudder) or EVPN multihoming
                "peer_link": False,  # don't care about MC-LAG (shudder)
            }

        # adding children to parents
        for name, entry in children.items():

            # want physical interface names only
            name = name.split(".")[0]
            lag_name = entry.lag_name.split(".")[0]

            if lag_name not in lag_interfaces:
                raise UMnetNapalmError(f"Invalid LAG assignment {lag_name} for {name}")

            lag_interfaces[lag_name]["members"][name] = {
                "oper_up": entry.oper_status == "up",
                "admin_up": entry.admin_status == "up"
                and entry.phy_admin_status == "up",
                "flags": "",
            }

        # next we're going to flag LACP ports
        try:
            result = self._send_command("junos_show_lacp")

        # if lacp isn't running the junos device returns an RPC error,
        # which raises an exception
        except Exception as e:

            # if the exception is because lacp isn't running, return
            # our list of LAGs as is, otherwise we want to raise
            if "lacp subsystem not running" in e.message:
                return lag_interfaces
            raise

        for lag_name, entry in dict(result).items():

            for member_name, member_entry in dict(entry.members).items():
                if (
                    member_name
                    not in lag_interfaces.get(lag_name, {"members": {}})["members"]
                ):
                    raise UMnetNapalmError(
                        f"Invalid LACP entry for LAG {lag_name} member {member_name}"
                    )

                # all LAGs appear in 'show lacp' but the ones actually running LACP have members.
                lag_interfaces[lag_name]["protocol"] = "LACP"

                # lacp_activity = "Active" or "Passive", lacp_timeout = "Slow" or "Fast"
                flags = member_entry.lacp_activity[0] + member_entry.lacp_timeout[0]
                lag_interfaces[lag_name]["members"][member_name]["flags"] = flags

        return lag_interfaces



import xml.etree.ElementTree as ET
from typing import List, Dict
import re

import xmltodict

from napalm.base import NetworkDriver
import napalm.base.models as napalm_models
from napalm.base.exceptions import ConnectionException
import pan.xapi

from umnet_napalm.models import (
    IPInterfaceDict,
    RouteDict,
    InventoryDict,
    LagInterfaceDict,
)
from ..base import UMnetNapalm, UMnetNapalmError
from .utils import parse_system_state
from ..utils import age_to_integer


PANOS_ROUTING_PROTOCOLS = {
    "H": "local",
    "C": "connected",
    "S": "static",
    "B": "BGP",
    "R": "RIP",
    "O": "OSPF",
    "Oi": "OSPF intra-area",
    "Oo": "OSPF inter-area",
    "O2": "OSPF External Type 2",
    "O1": "OSPF External Type 1",
}


class PANOS(UMnetNapalm, NetworkDriver):
    """
    PANOS Parser
    """

    # maps the 'type' field in slots/entry returned by 'show chassis inventory'
    # to our standard types
    INVENTORY_TO_TYPE = {
        "power": "psu",
        "slot": "linecard",
        "fan": "fan",
    }

    def __init__(self, hostname, username, password, timeout=60, optional_args=None):
        self.hostname = hostname
        self.username = username
        self.password = password
        self.timeout = timeout

        self.device = None

        self.api_key = optional_args.get("api_key", "")
        self._effective_running_config = None
        self._system_state = None

        self.serial = None
        if optional_args:
            self.serial = optional_args.get("serial")
            self.panorama = optional_args.get("panorama")

    def open(self):
        """PANOS version of `open` method, see NAPALM for documentation."""
        try:
            hostname = self.panorama if self.panorama else self.hostname

            if self.api_key:
                self.device = pan.xapi.PanXapi(
                    hostname=hostname, api_key=self.api_key, serial=self.serial
                )
            else:
                self.device = pan.xapi.PanXapi(
                    hostname=hostname,
                    api_username=self.username,
                    api_password=self.password,
                    serial=self.serial,
                )

        except ConnectionException as exc:
            raise ConnectionException(str(exc))

    def close(self):
        """PANOS version of `close` method, see NAPALM for documentation."""
        self.device = None

    def _config_search(self, xpath: str, no_cache: bool = False) -> list:
        """
        Does an xpath search of the effective running config and returns the results.
        """
        cfg = self._get_effective_running_config(no_cache=no_cache)
        return cfg.findall(xpath)

    def _get_effective_running_config(self, no_cache: bool = False) -> ET:
        """
        Gets effective running config and saves it to the local object
        """
        if self._effective_running_config is None or no_cache:
            result = self._send_command(
                "<show><config><effective-running></effective-running></config></show>"
            )
            self._effective_running_config = ET.fromstring(result)

        return self._effective_running_config

    def _get_system_state(self, no_cache: bool = False) -> dict:
        """
        runs 'show system state' and parses the returned pseudo-json into structured output
        """
        if not self._system_state or no_cache:
            raw_result = self._send_command("<show><system><state/></system></show>")
            result = xmltodict.parse(raw_result)["response"]["result"]
            self._system_state = parse_system_state(result)

        return self._system_state

    def _get_mtu_and_comment(self, interface: ET) -> dict:
        """
        Looks up mtu and comment on a particular interface or subinterface
        object. Returns a dict with mtu and comment key/value pairs
        """
        results = {"mtu": 0, "comment": ""}

        # subinterfaces have mtu set at their same level.
        # for non-subinterfaces it is set under "layer3"
        if "." in interface.attrib["name"]:
            mtu = interface.find("./mtu")
        else:
            mtu = interface.find("./layer3/mtu")

        if mtu is not None:
            results["mtu"] = mtu.text

        comment = interface.find("./comment")
        if comment is not None:
            results["comment"] = comment.text

        return results

    def _get_interface_mtus_and_comments(self) -> dict:
        """
        Looks through the effective running config at network/interface for MTU configurations
        and descriptions, aka "comments"
        returns a dict keyed on interface names, with
        dict vaules, eg results[interface_name] = {'mtu':mtu, and 'comment': comment }
        """

        interfaces = self._config_search(".//network/interface/ethernet/entry")

        # for logical interfaces we only care about the subints
        interfaces.extend(
            self._config_search(".//network/interface/tunnel/units/entry")
        )
        interfaces.extend(
            self._config_search(".//network/interface/loopback/units/entry")
        )

        results = {}
        # looping over all the ethernet interfaces
        for interface in interfaces:
            results[interface.attrib["name"]] = self._get_mtu_and_comment(interface)

            # subinterfaces
            for subint in interface.findall("./layer3/units/entry"):
                results[subint.attrib["name"]] = self._get_mtu_and_comment(subint)

        return results

    def _send_command(self, cmd: str) -> str:
        """
        Send XML command to PANOS and get result
        """
        self.device.op(cmd=cmd)
        return self.device.xml_root()

    def get_facts(self) -> napalm_models.FactsDict:
        """
        Parses 'show system info'
        """
        result = self._send_command("<show><system><info/></system></show>")
        sys_info = xmltodict.parse(result)["response"]["result"]["system"]

        return {
            "os_version": sys_info["sw-version"],
            "uptime": age_to_integer(sys_info["uptime"]),
            "interface_list": [],  # we don't care about this
            "vendor": "Palo Alto",
            "serial_number": sys_info["serial"],
            "model": sys_info["model"],
            "hostname": sys_info["hostname"],
            "fqdn": "NA",
        }

    def get_config(self, retrieve="all", full=False, sanitized=False):
        """
        We just want the effective running config
        """
        return self._send_command(
            "<show><config><effective-running></effective-running></config></show>"
        )

    def get_inventory(self) -> List[InventoryDict]:
        """
        Digs through "show chassis inventory" and "show system state" to pull out linecard and optics information
        """
        # need access to a chassis PAN to implement "show chassis inventory"
        result = self._send_command("<show><chassis><inventory/></chassis></show>")
        inventory = ET.fromstring(result)

        output = []

        # modules
        for entry in inventory.findall(".//slots/entry"):
            if entry.find("model").text in ["empty", "Not Present"]:
                continue

            inventory_type = self._get_inventory_type(entry.find("type").text)
            if not inventory_type:
                continue

            name = entry.find("slot").text
            name = re.sub(r"^(\d+)$", r"Slot \g<1>", name)

            output.append(
                {
                    "type": inventory_type,
                    "name": name,
                    "part_number": entry.find("model").text,
                    "serial_number": entry.find("serial").text,
                }
            )

        # optics data is in system state
        raw_result = self._send_command(
            "<show><system><state><filter>sys.s*.p*.phy</filter></state></system></show>"
        )
        data = parse_system_state(raw_result)

        for phy_name, phy_data in data.items():

            # skip unpopulated ports
            # if re.match(r'(CAT|-Empty$)', phy_data["media"]):
            if not phy_data.get("sfp"):
                continue

            # the 'phy name' is in the format 'sys.sX.pY.phy' which translates to 'EthernetX/Y'
            m = re.match(r"sys.s(\d+).p(\d+).phy$", phy_name)
            i_name = f"Ethernet{m.group(1)}/{m.group(2)}"

            output.append(
                {
                    "type": "optic",
                    "name": i_name,
                    "part_number": phy_data["sfp"]
                    .get("vendor-part-number", "")
                    .strip(),
                    "serial_number": phy_data["sfp"]
                    .get("vendor-serial-number", "")
                    .strip(),
                    "parent": "",
                }
            )

        return output

    def get_arp_table(self) -> List[napalm_models.ARPTableDict]:
        """
        Queries the arp table
        """
        result = self._send_command("<show><arp><entry name = 'all'/></arp></show>")
        arp_table = xmltodict.parse(result)["response"]["result"]

        timeout = int(arp_table["timeout"])

        output = []
        for entry in arp_table["entries"]["entry"]:
            output.append(
                {
                    "interface": entry["interface"],
                    "mac": entry["mac"],
                    "ip": entry["ip"],
                    "age": timeout - int(entry["ttl"]),
                },
            )

        return output

    def get_ip_interfaces(self) -> List[IPInterfaceDict]:
        """
        napalm_panos doesn't support "get_routing_instances".
        While we could implement that function and combine it with "get_interfaces_ip" and
        "get_interfaces", it's simpler to just run "show interface all" to get most of what we need.

        Note that MTU is in 'system state' so we have to parse it out of there.
        """

        # "show interface all" gives us most of what we need here
        result = self._send_command("<show><interface>all</interface></show>")
        interfaces = xmltodict.parse(result)["response"]["result"]

        hw_interfaces_by_name = {i["name"]: i for i in interfaces["hw"]["entry"]}
        mtus_and_comments = self._get_interface_mtus_and_comments()

        l3_interfaces = []
        for l3_i in interfaces["ifnet"]["entry"]:
            if l3_i["ip"] == "N/A" and not l3_i["addr6"]:
                continue

            # vrf (aka route-domain) is stored in the 'fwd' attribute
            vrf = None
            if l3_i["fwd"] != "N/A":
                vrf = l3_i["fwd"].replace("vr:", "")

            # look up admin/oper status in 'hw' section of the output
            if "." in l3_i["name"]:
                phy_name = l3_i["name"].split(".")[0]
            else:
                phy_name = l3_i["name"]
            phy = hw_interfaces_by_name.get(phy_name, None)

            # description and mtu came from an xpath query
            mtu_and_desc = mtus_and_comments.get(
                l3_i["name"], {"mtu": 0, "comment": ""}
            )

            # all the non-IP attributes for each IP entry are the same
            l3i_entry = {
                "interface": l3_i["name"],
                "description": mtu_and_desc["comment"],
                "mtu": mtu_and_desc["mtu"] if mtu_and_desc["mtu"] else 1500,
                "admin_up": bool(phy["mode"] != "(power-down)") if phy else None,
                "oper_up": bool(phy["state"] == "up") if phy else None,
                "vrf": vrf,
                # no helpers or secondaries - not currently configured
                # in our environment
                "secondary": False,
                "helpers": [],
            }

            # IPv4 interfaces
            if l3_i["ip"] and l3_i["ip"] != "N/A":
                ipv4_entry = l3i_entry.copy()
                ipv4_entry["ip_address"] = l3_i["ip"]
                l3_interfaces.append(ipv4_entry)

            # IPv6 interfaces are stored as a list, or a string if there's only one
            if l3_i["addr6"]:
                if isinstance(l3_i["addr6"]["member"], str):
                    l3_i["addr6"]["member"] = [l3_i["addr6"]["member"]]

                for ipv6_addr in l3_i["addr6"]["member"]:
                    ipv6_entry = l3i_entry.copy()
                    ipv6_entry["ip_address"] = ipv6_addr
                    l3_interfaces.append(ipv6_entry)

        return l3_interfaces

    def get_active_routes(self) -> List[RouteDict]:
        """
        Gets active routes from all vsyses on the PAN
        """
        result = self._send_command("<show><routing><route/></routing></show>")
        routing = xmltodict.parse(result)

        parsed_routes = []
        for route in routing["response"]["result"]["entry"]:

            # if route isn't active (no A in protocol field), skip it
            if not (route["flags"].startswith("A")):
                continue

            # look at the third character in the flags field to determine
            # protocol - don't really care about the different OSPF types
            if route["flags"][2] not in PANOS_ROUTING_PROTOCOLS:
                raise UMnetNapalmError(
                    f"{self.hostname}: Unknown panos routing protocol {route['flags']}"
                )
            protocol = PANOS_ROUTING_PROTOCOLS[route["flags"][2]]

            # next hops of all zeros indicate local routes
            if route["nexthop"] in ["0.0.0.0", "::"]:
                learned_from = "self"
            else:
                learned_from = route["nexthop"]

            # if this is a local route and we don't have a nh interface, the
            # next hop interface is also 'self' (because it's a host route or a locally-owned IP)
            if learned_from == "self" and not route["interface"]:
                route["interface"] = "self"

            parsed_routes.append(
                {
                    "vrf": route["virtual-router"],
                    "prefix": route["destination"],
                    "nh_interface": route["interface"],
                    "nh_table": route["virtual-router"],
                    "learned_from": learned_from,
                    "protocol": protocol,
                    "age": int(route["age"]) if route["age"] else None,
                    "nh_ip": None if learned_from == "self" else route["nexthop"],
                    "mpls_label": [],
                    "vxlan_vni": None,
                    "vxlan_endpoint": None,
                }
            )

        return parsed_routes

    def get_lag_interfaces(self) -> Dict[str, LagInterfaceDict]:
        """
        parses "show interface interface" and "show lacp aggregate-ethernet all"
        """
        result = self._send_command(
            "<show><lacp><aggregate-ethernet>all</aggregate-ethernet></lacp></show>"
        )
        lacp = ET.fromstring(result)

        lag_interfaces = {}
        for entry in lacp.findall("result/entry"):
            lag_name = entry.attrib.get("name")
            if not lag_name:
                continue

            # 'mode' is Active or Passive, 'rate' is Slow or Fast, and is set
            # at the parent level
            flags = entry.find("mode").text[0] + entry.find("rate").text[0]

            # to get oper status of parent we are going to parse the
            # the members first
            lag_members = {}
            for member in entry.findall("entry"):
                lag_members[member.attrib["name"]] = {
                    "oper_up": member.find("lacp-state") == "active",
                    "admin_up": True,
                    "flags": flags,
                }

            lag_interfaces[lag_name] = {
                # if one of the members is up then the parent is up
                "oper_up": len(
                    [m["oper-up"] for m in lag_members.values() if m["oper_up"]]
                ),
                "admin_up": entry.find("lacp").text == "Enabled",
                "protocol": "LACP",  # only supporting LACP
                "peer_link": False,
                "mlag_id": 0,
                "members": lag_members,
            }

        return lag_interfaces



import re
from typing import List, Dict
from ipaddress import ip_network
import logging

from napalm.ios import IOSDriver
from napalm.base.helpers import textfsm_extractor

from ..models import (
    RouteDict,
    MPLSDict,
    InventoryDict,
    IPInterfaceDict,
    LagInterfaceDict,
)
from ..base import UMnetNapalmError, UMnetNapalm
from ..utils import age_to_integer

IOS_LABEL_VALUES = {
    "": [],
    "No Label": [],
    "Pop Label": ["pop"],
}

logger = logging.getLogger(__name__)


class IOS(UMnetNapalm, IOSDriver):
    """
    IOS Parser
    """

    LABEL_VALUE_MAP = {
        "": [],
        "No Label": [],
        "Pop Label": ["pop"],
    }

    INVENTORY_TO_TYPE = {
        # weird one-offs we care about
        r"4500X-16 10GE": "stack_member",  # umd core vss "stack_member"
        r"10/100BaseTX \(RJ45\)V? with 48 10/100 baseT": "linecard",  # linecard in 4500 chassis
        # re
        r"Supervisor": "re",
        # linecard
        r"^WS-X": "linecard",
        r"Async Serial NIM": "linecard",  # ISR NIM
        r"Eth Mod": "linecard",
        r"HWIC": "linecard",
        r"Fan": "fan",
        r"[Pp]ower [Ss]upply": "psu",
        r"RPS": "psu",
        r"Uplink Module": "uplink_module",
        r"FRULink": "uplink_module",
        r"gigabit ethernet port adapter": "uplink_module",
        r"(WS-C[234]|C9[23]00)": "stack_member",
        r"StackPort": "stack_cable",
        r"StackAdapter": "stack_cable",
        # optic matches are last since they're loose
        r"Interface Adapter": "optic",
        r"TwinGig": "optic",
        r"Converter Module": "optic",
        r"Transceiver": "optic",
        r"SFP": "optic",
        r"1000[Bb]ase": "optic",
        r"10G[Bb]ase": "optic",
        # items we want to ignore
        r"WS-F": None,  # forwarding card
        r"Daughterboard": None,  # daughterboard
        r"Policy Feature Card": None,
        r"[Cc]hassis": None,
        r"Clock": None,
        r"Mux Buffers": None,
        r"c(92|93|36)xx": None,  # duplicate 'chassis' entry
        r"C1000": None,  # AV C1000 chassis
        r"Cisco [Cc]atalyst c2950 switch": None,  # s-vaughn-252c-1 c2950
        r"Cisco [Cc]atalyst 3550": None,  # s-nccomm-1241t-1
        r"Cisco Systems, Inc. WS-C4500X-16 2 slot switch": None,  # umd core vss sup entry
        r"Stacking Module": None,  # replaceable stacking module in 2960X
        # terminal server components we don't care about
        r"terminal server": None,
        r"Built-In NIM controller": None,
        r"Front Panel 2 ports Gigabitethernet Module": None,
        r"(Route|Forwarding) Processor": None,
    }

    I_ABBRS = {
        "Lo": "Loopback",
        "Po": "Port-channel",
        "Fa": "FastEthernet",
        "Ge": "GigabitEthernet",
        # longer matches must go first!
        "Twe": "TwentyFiveGigE",
        "Tw": "TwoGigabitEthernet",
        "Te": "TenGigabitEthernet",
        "Vl": "Vlan",
    }

    PROTOCOL_ABBRS = {
        "L": "local",
        "C": "connected",
        "S": "static",
        "i": "ISIS",
        "B": "BGP",
        "O": "OSPF",
    }

    def __init__(self, hostname, username, password, timeout=60, optional_args=None):

        if not optional_args:
            optional_args = {}
        optional_args["read_timeout_override"] = timeout

        super().__init__(hostname, username, password, timeout, optional_args)

        self._i_descrs = None

    def _parse_label_value(self, label) -> list:
        """
        Parses mpls label value into normalized data
        """
        if label in IOS_LABEL_VALUES:
            return IOS_LABEL_VALUES[label]

        return [label]

    def _get_route_labels(self) -> Dict[tuple, str]:
        """
        Runs "show bgp vpnv4 unicast labels" and parses the result.

        The output is a dictionary with (vrf, prefix) as key
        and the outbound label as a value, eg
        output[ ("vrf_VOIP_NGFW", "0.0.0.0/0") ] = "12345"
        """

        raw_labels = self._send_command("show bgp vpnv4 unicast all labels")
        labels = textfsm_extractor(self, "sh_bgp_vpnv4_unicast_all_labels", raw_labels)

        # default route shows up as '0.0.0.0' so we have to munge that
        output = {}
        for l in labels:
            prefix = "0.0.0.0/0" if l["prefix"] == "0.0.0.0" else l["prefix"]
            output[(l["vrf"], ip_network(prefix))] = l["out_label"]

        return output

    @property
    def i_descrs(self):
        """
        Pulls interface descriptions into an internal data structure
        """

        if self._i_descrs is None:
            raw_output = self._send_command("show interface description")
            output = textfsm_extractor(self, "sh_interface_description", raw_output)
            self._i_descrs = {o["interface"]: o["description"] for o in output}
        return self._i_descrs

    def get_ip_interfaces(self) -> List[IPInterfaceDict]:
        """
        Parses 'show ip interface' and 'show ipv6 interface'
        """
        raw_output = self._send_command("show ip interface")
        output = textfsm_extractor(self, "sh_ip_interface", raw_output)

        ip_interfaces = []
        for i in output:

            i_data = {
                "ip_address": i["ip"],
                "interface": i["interface"],
                "description": self.i_descrs.get(self._abbr_i(i["interface"]), ""),
                "mtu": int(i["mtu"]) if i["mtu"] else None,
                "admin_up": i["admin_state"] != "administratively down",
                "oper_up": i["oper_state"] == "up",
                "vrf": i["vrf"] if i["vrf"] else "default",
                "secondary": False,
                "helpers": set(i["helpers"]) if i["helpers"] else [],
            }
            ip_interfaces.append(i_data)

            for sec_i in i["sec_ip"]:

                sec_data = i_data.copy()
                sec_data["ip_address"] = sec_i
                sec_data["secondary"] = True

                ip_interfaces.append(sec_data)

        raw_output = self._send_command("show ipv6 interface")
        output = textfsm_extractor(self, "sh_ipv6_interface", raw_output)

        for i in output:
            for ip in i["ip"]:
                ip_interfaces.append(
                    {
                        "ip_address": ip,
                        "interface": i["interface"],
                        "description": self.i_descrs.get(
                            self._abbr_i(i["interface"]), ""
                        ),
                        "mtu": int(i["mtu"]) if i["mtu"] else None,
                        "admin_up": i["admin_state"] != "administratively down",
                        "oper_up": i["oper_state"] == "up",
                        "vrf": i["vrf"] if i["vrf"] else "default",
                        # IPv6 secondary and helpers not relevant in our environment
                        "secondary": False,
                        "helpers": [],
                    }
                )

        return ip_interfaces

    def get_active_routes(self) -> List[RouteDict]:
        """
        Parses "show ip route vrf *" for IOS. Will also run
        "show bgp vpnv4 unicast labels" to get label bindings
        """

        output = []

        raw_routes = self._send_command("show ip route vrf *")
        routes = textfsm_extractor(self, "sh_ip_route_vrf_all", raw_routes)

        for route in routes:
            logger.info(f"found proto {route['proto_1']} for route {route}")
            protocol = self._parse_protocol_abbr(route["proto_1"])

            # "learned from" is one of the keys in our route table that determines
            # uniqueness, as such we need to make sure it's set. Usually it's
            # the IP of the advertising router, but for local/direct/static it should
            # get set to 'self'
            if route["nh_ip"]:
                learned_from = route["nh_ip"]
            elif protocol in ["local", "connected", "static"]:
                learned_from = "self"
            else:
                raise UMnetNapalmError(f"Could not determine learned from for {route}")

            output.append(
                {
                    "vrf": route["vrf"] if route["vrf"] else "default",
                    "prefix": route["prefix"],
                    "nh_interface": route["nh_interface"],
                    "learned_from": learned_from,
                    "protocol": self._parse_protocol_abbr(route["proto_1"]),
                    "age": age_to_integer(route["age"]),
                    "nh_ip": route["nh_ip"],
                    "mpls_label": None,
                    "vxlan_vni": None,
                    "vxlan_endpoint": None,
                    "nh_table": route["vrf"] if route["vrf"] else "default",
                }
            )

        return output

    def get_inventory(self) -> list[InventoryDict]:
        """
        Parses "show inventory" for IOS
        """
        raw_inventory = self._send_command("show inventory")
        inventory = textfsm_extractor(self, "sh_inventory", raw_inventory)

        output = []
        for entry in inventory:
            inventory_type = self._get_inventory_type(entry["desc"])

            if not inventory_type:
                continue

            name = re.sub(r"^(\d+)$", r"Slot \g<1>", entry["name"])

            output.append(
                {
                    "name": name,
                    "type": inventory_type,
                    "part_number": entry["pid"],
                    "serial_number": entry["sn"],
                }
            )

        return output

    def get_mpls_switching(self) -> List[MPLSDict]:
        """
        Parses "show mpls forwarding table" for IOS
        """
        raw_labels = self._send_command("show mpls forwarding-table")
        labels = textfsm_extractor(self, "sh_mpls_forwarding_table", raw_labels)

        output = []
        for entry in labels:
            # extract RD from 'FEC'
            m = re.match(r"([\d\.]+:\d+):(\d+.\d+.\d+.\d+\/\d+)", entry["fec"])
            if m:
                rd = m.group(1)
                fec = m.group(2)
            else:
                rd = None
                fec = entry["fec"]

            aggregate = bool(entry["vrf"])
            nh_interface = entry["vrf"] if entry["vrf"] else entry["nh_interface"]

            output.append(
                {
                    "in_label": entry["in_label"],
                    "fec": fec,
                    "out_label": self._parse_label_value(entry["out_label"]),
                    "nh_ip": entry["nh_ip"],
                    "nh_interface": nh_interface,
                    "rd": rd,
                    "aggregate": aggregate,
                }
            )

        return output

    def get_lag_interfaces_old(self) -> Dict[str, LagInterfaceDict]:
        """
        Parses "show etherchannel detail"
        """
        raw_output = self._send_command("show etherchannel detail")
        output = textfsm_extractor(self, "sh_etherchannel_detail", raw_output)

        # template has two distinct sets of fields, one for lags
        # and the other for members. First we'll initialize our parent class
        groups = [line for line in output if line["lag_name"]]
        members = [line for line in output if line["member_name"]]

        lag_interfaces = {}
        for line in groups:
            lag_interfaces[line["lag_name"]] = {
                "admin_up": "Down" not in line["lag_state"],
                "oper_up": "Ag-Inuse" in line["lag_state"],
                "protocol": line["lag_proto"] if line["lag_proto"] != "-" else "Static",
                "mlag_id": 0,  # no vpc/mlag for ios
                "members": {},
            }

        for line in members:
            group_name = f"Po{line['member_group']}"
            if group_name not in lag_interfaces:
                raise UMnetNapalmError(
                    f"Unknown LAG group {group_name} for member {line['member_name']}"
                )

            lag_interfaces[group_name]["members"][line["member_name"]] = {
                "admin_up": True,  # don't see member admin state in 'show etherchannel detail'
                "oper_up": line["member_state"] == "Up",
                "flags": line["member_flags"].strip(),
            }

        return lag_interfaces

    def get_lag_interfaces(self) -> Dict[str, LagInterfaceDict]:
        """
        Parses "show etherchannel summary"
        """
        raw_output = self._send_command("show etherchannel summary")
        output = textfsm_extractor(self, "sh_etherchannel_summary", raw_output)

        lag_interfaces = {}
        for line in output:
            if line["lag_name"] not in lag_interfaces:
                lag_interfaces[line["lag_name"]] = {
                    "oper_up": "D" not in line["lag_flags"],
                    "admin_up": True,  # don't see member admin state in 'show etherchannel summ'
                    "protocol": (
                        line["protocol"] if line["protocol"] != "-" else "Static"
                    ),
                    "mlag_id": 0,
                    "peer_link": False,
                    "members": {},
                }

            lag = lag_interfaces[line["lag_name"]]

            for member in ["m1", "m2"]:

                member_name = line[f"{member}_name"]
                member_flags = line[f"{member}_flags"]
                if not member_name:
                    continue

                lag["members"][member_name] = {
                    "oper_up": "P" in member_flags,
                    "admin_up": True,
                    "flags": member_flags,
                }

        return lag_interfaces



from typing import List
import logging

from lxml import etree

from napalm.iosxr_netconf import IOSXRNETCONFDriver
from ncclient.xml_ import to_ele

from ..base import UMnetNapalm
from ..models import RouteDict, IPInterfaceDict, InventoryDict

from .constants import IP_INT_RPC_REQ, IP_ROUTE_RPC_REQ, NS, INV_RPC_REQ

logger = logging.getLogger(__name__)


class IOSXRNetconf(IOSXRNETCONFDriver, UMnetNapalm):
    """
    IOSXR Class
    """

    INVENTORY_TO_TYPE = {
        r"PM\d+$": "psu",
        r"FT\d+$": "fan",
        r"FC\d+$": "fabric_module",
        r"RP\d$": "re",
        r"SC\d": None,  # don't care about system controller
        r"Rack 0": None,  # this is the chassis, don't care about it
        r"^(FourHundred|Hundred|Forty)GigE": "optic",
        r"\d/\d": "linecard",
    }

    # Helper xml methods that always pass in our namespaces by default
    def _text(self, xml_tree, path, default=None, namespaces=NS):
        return super()._find_txt(xml_tree, path, default, namespaces=namespaces)

    def _xpath(self, xml_tree, path, namespaces=NS):
        return getattr(xml_tree, "xpath")(path, namespaces=namespaces)

    def _find(self, xml_tree, element, namespaces=NS):
        return getattr(xml_tree, "find")(element, namespaces=namespaces)

    def _iterfind(self, xml_tree, element, namespaces=NS):
        return getattr(xml_tree, "iterfind")(element, namespaces=namespaces)

    def get_ip_interfaces(self) -> List[IPInterfaceDict]:
        """
        Gets IP interfaces
        """
        rpc_reply = self.device.dispatch(to_ele(IP_INT_RPC_REQ)).xml
        xml_result = etree.fromstring(rpc_reply)
        results = []

        # looking at every interface and pulling operation state and description
        for i in self._xpath(xml_result, "//int:interfaces/int:interface"):

            i_name = self._text(i, "int:interface-name")
            i_data = {
                "interface": i_name,
                "description": self._text(i, "int:description"),
                "admin_up": self._text(i, "int:state") != "im-state-admin-down",
                "oper_up": self._text(i, "int:line-state") == "im-state-up",
            }

            # pulling primary IPv4 for the interface. Note that though we're looping,
            # we only expect the xpath query to return a single entry.
            for pri_ipv4 in self._xpath(
                xml_result,
                f"//int4:vrf/int4:details/int4:detail[int4:interface-name='{i_name}']",
            ):
                ip = self._text(pri_ipv4, "int4:primary-address")

                prefixlen = self._text(pri_ipv4, "int4:prefix-length")

                i_data.update(
                    {
                        "vrf": self._text(pri_ipv4, "../../int4:vrf-name"),
                        "mtu": self._text(pri_ipv4, "int4:mtu"),
                    }
                )
                if ip != "0.0.0.0":
                    i_data.update(
                        {
                            "ip_address": f"{ip}/{prefixlen}",
                            "mtu": self._text(pri_ipv4, "int4:mtu"),
                        }
                    )
                    results.append(i_data.copy())

                # ipv4 secondaries
                for sec_ipv4 in self._iterfind(pri_ipv4, "int4:secondary-address"):
                    ip = self._text(sec_ipv4, "int4:address")
                    prefixlen = self._text(sec_ipv4, "int4:prefix-length")
                    i_data["ip_address"] = f"{ip}/{prefixlen}"
                    results.append(i_data.copy())

            # ipv6 addresses
            for ipv6 in self._xpath(
                xml_result,
                f"//int6:global-detail[int6:interface-name='{i_name}']/int6:address",
            ):
                i_data["mtu"] = self._text(ipv6, "../int6:mtu")
                ip = self._text(ipv6, "int6:address")
                prefixlen = self._text(ipv6, "int6:prefix-length")
                i_data["ip_address"] = f"{ip}/{prefixlen}"
                results.append(i_data.copy())

        return results

    def get_active_routes(self) -> List[RouteDict]:
        """
        Pulls active routes from the rib
        """
        rpc_reply = self.device.dispatch(to_ele(IP_ROUTE_RPC_REQ)).xml
        xml_result = etree.fromstring(rpc_reply)

        output = []

        # ipv4 rib and ipv6 rib yang models are similar enough that we can
        # use the same logic for both
        for v in ["4", "6"]:
            for route in self._xpath(
                xml_result,
                f"//rib{v}:ip-rib-route-table-name/rib{v}:routes/rib{v}:route",
            ):
                vrf = self._text(route, f"../../../../../../../../rib{v}:vrf-name")

                subnet = self._text(route, f"rib{v}:address")
                prefixlen = self._text(route, f"rib{v}:prefix-length")
                protocol = self._text(route, f"rib{v}:protocol-name")
                age = self._text(route, f"rib{v}:route-age")

                for nh in self._xpath(
                    route, f"rib{v}:route-path/rib{v}:ipv{v}-rib-edm-path"
                ):
                    learned_from = self._text(nh, f"rib{v}:information-source")
                    if learned_from == "0.0.0.0":
                        learned_from = "self"

                    # not currently dealing with enapsulation (where the next hop
                    # could be in a different table), but we do have some PBR
                    # that shows up under 'next-hop-vrf-name
                    if self._text(nh, f"rib{v}:next-hop-vrf-name"):
                        nh_table = self._text(nh, f"rib{v}:next-hop-vrf-name")
                    else:
                        nh_table = vrf

                    output.append(
                        {
                            "vrf": vrf,
                            "prefix": f"{subnet}/{prefixlen}",
                            "nh_interface": self._text(nh, f"rib{v}:interface-name"),
                            "learned_from": learned_from,
                            "protocol": protocol,
                            "nh_ip": self._text(nh, f"rib{v}:address"),
                            "age": age,
                            "mpls_label": [],
                            "vxlan_vni": None,
                            "nh_table": nh_table,
                        }
                    )

        return output

    def get_inventory(self) -> List[InventoryDict]:
        """
        Gets inventory data
        """
        rpc_reply = self.device.dispatch(to_ele(INV_RPC_REQ)).xml
        xml_result = etree.fromstring(rpc_reply)

        output = []
        for item in self._xpath(xml_result, "//inv:inv-basic-bag"):

            name = self._text(item, "inv:name")
            item_type = self._get_inventory_type(name)
            if item_type:
                output.append(
                    {
                        "type": item_type,
                        "name": name,
                        "part_number": self._text(item, "inv:model-name"),
                        "serial_number": self._text(item, "inv:serial-number"),
                        "parent": None,
                    }
                )

        return output



from concurrent.futures import ProcessPoolExecutor
from typing import List

from nornir.core.inventory import Host
from nornir.core.task import AggregatedResult, Task


class MultiProcessRunner:
    """
    MultiProcessRunner runs the task over each host using multiple processors

    Arguments:
        num_workers: number of processors to use
    """

    def __init__(self, num_workers: int = 10) -> None:
        self.num_workers = num_workers

    def run(self, task: Task, hosts: List[Host]) -> AggregatedResult:
        result = AggregatedResult(task.name)
        futures = []
        with ProcessPoolExecutor(self.num_workers) as pool:
            for host in hosts:
                future = pool.submit(task.copy().start, host)
                futures.append(future)

        for future in futures:
            worker_result = future.result()
            result[worker_result.host.name] = worker_result
        return result
